package org.processmining.datapetrinets.syncproduct;

import java.text.ParseException;
import java.util.HashMap;
import java.util.Map;

import org.processmining.framework.util.Pair;
import org.processmining.models.graphbased.LocalNodeID;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Arc;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithData;
import org.processmining.models.guards.Expression;

public class DPNSynchronousProductComputation {

	/**
	 * TODO: Write Javadoc
	 * 
	 * @param sourceNetA
	 * @param sourceNetB
	 * @param netName
	 * @return
	 */
	public static PetriNetWithData computeSynchronousProduct(PetriNetWithData sourceNetA, PetriNetWithData sourceNetB, String netName) {
		PetriNetWithData targetNet = new PetriNetWithData(netName);
		Map<LocalNodeID,Place> sourceNetAPlaces = DPNUtility.getPlaceMap(sourceNetA);
		Map<LocalNodeID,Place> sourceNetBPlaces = DPNUtility.getPlaceMap(sourceNetB);
		Map<LocalNodeID,Place> targetNetPlaces = new HashMap<LocalNodeID,Place>();
		Map<LocalNodeID,Transition> sourceNetATransitions = DPNUtility.getTransitionMap(sourceNetA);
		Map<LocalNodeID,Transition> sourceNetBTransitions = DPNUtility.getTransitionMap(sourceNetB);
		Map<String,Pair<Transition,Transition>> synchronousTransitions = new HashMap<String,Pair<Transition,Transition>>();
		
		/** Find the overlap in transitions in sourceNetA and sourceNetB */
		for (Transition sourceTransitionA : sourceNetA.getTransitions()) {
			for (Transition sourceTransitionB : sourceNetB.getTransitions()) {
				if (sourceTransitionA.getLabel().equals(sourceTransitionB.getLabel())) {
					sourceNetATransitions.remove(sourceTransitionA.getLocalID());
					sourceNetBTransitions.remove(sourceTransitionB.getLocalID());
					synchronousTransitions.put(sourceTransitionA.getLabel(),new Pair<Transition,Transition>(sourceTransitionA,sourceTransitionB));
				}
			}
		}
		
		/** Add all unmatched transitions from sourceNetA to targetNet */
		for (Transition sourceTransition : sourceNetATransitions.values()) {
			PNWDTransition targetTransition = (PNWDTransition) targetNet.addTransition(sourceTransition.getLabel());
			targetTransition.setLocalID(sourceTransition.getLocalID());
			targetTransition.setInvisible(sourceTransition.isInvisible());
			
			/** Add all edges of sourceTransition to targetTransition. */
			addEdgesAndPlacesForTransition(sourceNetA, targetNet, sourceTransition, targetTransition, sourceNetAPlaces,
					targetNetPlaces);
			
			/** Add all variables, read & write operations, and guards. */
			addVariableReadWrite(targetNet, sourceTransition, targetTransition);
			if (((PNWDTransition) sourceTransition).getGuard() != null) targetTransition.setGuard(targetNet, ((PNWDTransition) sourceTransition).getGuard());
		}
		
		/** Add all unmatched transitions from sourceNetB to targetNet */
		for (Transition sourceTransition : sourceNetBTransitions.values()) {
			PNWDTransition targetTransition = (PNWDTransition) targetNet.addTransition(sourceTransition.getLabel());
			targetTransition.setLocalID(sourceTransition.getLocalID());
			targetTransition.setInvisible(sourceTransition.isInvisible());
			
			/** Add all edges of sourceTransition to targetTransition. */
			addEdgesAndPlacesForTransition(sourceNetB, targetNet, sourceTransition, targetTransition, sourceNetBPlaces,
					targetNetPlaces);
			
			/** Add all variables, read & write operations, and guards. */
			addVariableReadWrite(targetNet, sourceTransition, targetTransition);
			if (((PNWDTransition) sourceTransition).getGuard() != null) targetTransition.setGuard(targetNet, ((PNWDTransition) sourceTransition).getGuard());
		}
		
		/** Add all synchronous transitions to targetNet */
		for (String overlap : synchronousTransitions.keySet()) {
			Pair<Transition,Transition> pair = synchronousTransitions.get(overlap);
			Transition sourceTransitionA = pair.getFirst();
			Transition sourceTransitionB = pair.getSecond();
			PNWDTransition targetTransition = (PNWDTransition) targetNet.addTransition(overlap);
			targetTransition.setInvisible(sourceTransitionA.isInvisible());
			
			/** Add all edges of both sourceTransitions to targetTransition. */
			addEdgesAndPlacesForTransition(sourceNetA, targetNet, sourceTransitionA, targetTransition, sourceNetAPlaces,
					targetNetPlaces);
			addEdgesAndPlacesForTransition(sourceNetB, targetNet, sourceTransitionB, targetTransition, sourceNetBPlaces,
					targetNetPlaces);
			
			/** Add all variables, read & write operations, and guards. */
			addVariableReadWrite(targetNet, sourceTransitionA, targetTransition);
			addVariableReadWrite(targetNet, sourceTransitionB, targetTransition);
			Expression sourceTransitionAGuard = ((PNWDTransition) sourceTransitionA).getGuard();
			Expression sourceTransitionBGuard = ((PNWDTransition) sourceTransitionB).getGuard();
			if (sourceTransitionAGuard != null) {
				if (sourceTransitionBGuard != null) {
					// Combine guards
					try {
						targetTransition.setGuardAsString(targetNet, "(" + sourceTransitionAGuard + "&&" + sourceTransitionBGuard + ")");
					} catch (ParseException e) {
						e.printStackTrace();
					}
				}
				else {
					targetTransition.setGuard(targetNet, ((PNWDTransition) sourceTransitionA).getGuard());
				}
			}
			else if (sourceTransitionBGuard != null) {
				targetTransition.setGuard(targetNet, ((PNWDTransition) sourceTransitionB).getGuard());
			}
		}
		
		/** Add all unmatched places from sourceNetA and sourceNetB to targetNet */
		for (Place place : sourceNetAPlaces.values()) {
			Place newPlace = targetNet.addPlace(place.getLabel());
			newPlace.setLocalID(place.getLocalID());
		}
		
		for (Place place : sourceNetBPlaces.values()) {
			Place newPlace = targetNet.addPlace(place.getLabel());
			newPlace.setLocalID(place.getLocalID());
		}
		
		return targetNet;
	}
	
	/**
	 * TODO: Write Javadoc
	 * 
	 * @param sourceNet
	 * @param targetNet
	 * @param sourceTransition
	 * @param targetTransition
	 * @param sourceNetPlaces
	 * @param targetNetPlaces
	 */
	private static void addEdgesAndPlacesForTransition(PetriNetWithData sourceNet, PetriNetWithData targetNet,
			Transition sourceTransition, Transition targetTransition, Map<LocalNodeID, Place> sourceNetPlaces,
			Map<LocalNodeID, Place> targetNetPlaces) {
		
		/** Add all incoming edges of sourceTransition to targetTransition. */
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : sourceNet.getInEdges(sourceTransition)) {
			if (edge.getSource() instanceof Place) {
				Place source = targetNetPlaces.get(edge.getSource().getLocalID());
				if (source == null) {
					/** Add as new place. */
					source = targetNet.addPlace(edge.getSource().getLabel());
					source.setLocalID(edge.getSource().getLocalID());
					targetNetPlaces.put(source.getLocalID(), source);
					sourceNetPlaces.remove(edge.getSource().getLocalID());
				}
				
				Arc newArc = targetNet.addArc(source, targetTransition, ((Arc) edge).getWeight());
				newArc.setLocalID(edge.getLocalID());
			}
		}
		
		/** Add all outgoing edges of sourceTransition to targetTransition. */
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : sourceNet.getOutEdges(sourceTransition)) {
			if (edge.getTarget() instanceof Place) {
				Place target = targetNetPlaces.get(edge.getTarget().getLocalID()); 
				if (target == null) {
					/** Add as new place. */
					target = targetNet.addPlace(edge.getTarget().getLabel());
					target.setLocalID(edge.getTarget().getLocalID());
					targetNetPlaces.put(target.getLocalID(), target);
					sourceNetPlaces.remove(edge.getTarget().getLocalID());
				}
								
				Arc newArc = targetNet.addArc(targetTransition, target, ((Arc) edge).getWeight());
				newArc.setLocalID(edge.getLocalID());
			}
		}
		
		return;
	}
	
	/**
	 * TODO: Write Javadoc
	 * 
	 * @param targetNet
	 * @param sourceTransition
	 * @param targetTransition
	 */
	private static void addVariableReadWrite(PetriNetWithData targetNet, Transition sourceTransition, Transition targetTransition) {
		for (DataElement original : ((PNWDTransition) sourceTransition).getWriteOperations()) {
			DataElement writeOperation = targetNet.getVariable(original.getVarName());
			if (writeOperation == null) writeOperation = targetNet.addVariable(original.getVarName(),
					original.getType(), original.getMinValue(), original.getMaxValue());
			targetNet.assignWriteOperation(targetTransition, writeOperation);
		}
		for (DataElement original : ((PNWDTransition) sourceTransition).getReadOperations()) {
			DataElement readOperation = targetNet.getVariable(original.getVarName());
			if (readOperation == null) readOperation = targetNet.addVariable(original.getVarName(),
					original.getType(), original.getMinValue(), original.getMaxValue());
			targetNet.assignReadOperation(targetTransition, readOperation);
		}
		
	}
}
