package org.processmining.datapetrinets.syncproduct;

import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithData;
import org.processmining.models.semantics.petrinet.Marking;

@Plugin(
		name = "Compute Data Petri Net Synchronous Product",
		parameterLabels = { "Data Petri Nets" },
		returnLabels = { "Synchronous Product", "Initial Marking", "Final Marking" },
		returnTypes = { PetriNetWithData.class, Marking.class, Marking.class },
		mostSignificantResult = 1,
		categories = { PluginCategory.Analytics},
		help = "Compute the synchronous product of two or more Data Petri Nets and return the merged result.")
public class DPNSynchronousProductPlugin {
	
	/**
	 * TODO: Write Javadoc
	 * 
	 * @param context
	 * @param nets
	 * @return
	 */
	@UITopiaVariant(
			uiLabel = "Compute Data Petri Net Synchronous Product",
			affiliation = "Eindhoven University of Technology",
			author = "Maikel L. van Eck",
			email = "m.l.v.eck@tue.nl",
			pack = "DataPetriNets")
	@PluginVariant(variantLabel = "Compute Data Petri Net Synchronous Product", requiredParameterLabels = { 0 })
	public Object[] computeDPNSynchronousProduct(final UIPluginContext context, PetriNetWithData... nets) {
		
		/** Check for the existence of markings of the nets, note that we assume a single final marking per net! */
		for (int i = 0; i < nets.length; i++) {
			DPNUtility.checkMarkings(context, nets[i]);
		}
		
		/** Calculate the synchronous product of all input models with their initial and final markings. */
		PetriNetWithData net = nets[0];
		Marking initialMarking = DPNUtility.getInitialMarking(context, nets[0]);
		Marking finalMarking = DPNUtility.getFinalMarkings(context, nets[0])[0];
		for (int i = 1; i < nets.length; i++) {
			net = DPNSynchronousProductComputation.computeSynchronousProduct(net, nets[i], "Synchronous Product");
			initialMarking = DPNUtility.mergeMarkings(net, initialMarking, DPNUtility.getInitialMarking(context, nets[i]));
			finalMarking = DPNUtility.mergeMarkings(net, finalMarking, DPNUtility.getFinalMarkings(context, nets[i])[0]);
			context.addConnection(new InitialMarkingConnection(net, initialMarking));
			context.addConnection(new FinalMarkingConnection(net, finalMarking));
		}
		
		return new Object[] { net, initialMarking, finalMarking };
	}
	
	
}
