package org.processmining.datapetrinets.syncproduct;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.swing.JOptionPane;
import javax.swing.JPanel;

import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.framework.connections.Connection;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.connections.annotations.ConnectionObjectFactory;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.PluginExecutionResult;
import org.processmining.framework.plugin.PluginParameterBinding;
import org.processmining.framework.util.Pair;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.LocalNodeID;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithData;
import org.processmining.models.semantics.petrinet.Marking;

public class DPNUtility {

	/**
	 * 
	 * @param context
	 * @param net
	 */
	public static void checkMarkings(UIPluginContext context, PetrinetGraph net) {
		// check existence of initial marking
		try {
			InitialMarkingConnection initCon = context.getConnectionManager().getFirstConnection(
					InitialMarkingConnection.class, context, net);

			if (((Marking) initCon.getObjectWithRole(InitialMarkingConnection.MARKING)).isEmpty()) {
				JOptionPane
				.showMessageDialog(
						new JPanel(),
						"The initial marking is an empty marking. If this is not intended, remove the currently existing InitialMarkingConnection object and then use \"Create Initial Marking\" plugin to create a non-empty initial marking.",
						"Empty Initial Marking", JOptionPane.INFORMATION_MESSAGE);
			}
		} catch (ConnectionCannotBeObtained exc) {
			if (JOptionPane.YES_OPTION == JOptionPane.showConfirmDialog(new JPanel(),
					"No initial marking is found for this model. Do you want to create one?", "No Initial Marking",
					JOptionPane.YES_NO_OPTION, JOptionPane.INFORMATION_MESSAGE)) {
				createMarking(context, net, InitialMarkingConnection.class);
			}
		}

		// check existence of final marking
		try {
			context.getConnectionManager().getFirstConnection(FinalMarkingConnection.class, context, net);
		} catch (ConnectionCannotBeObtained exc) {
			if (JOptionPane.YES_OPTION == JOptionPane.showConfirmDialog(new JPanel(),
					"No final marking is found for this model. Do you want to create one?", "No Final Marking",
					JOptionPane.YES_NO_OPTION, JOptionPane.INFORMATION_MESSAGE)) {
				createMarking(context, net, FinalMarkingConnection.class);
			}
		} 
	}
	
	/**
	 * TODO: Write Javadoc
	 * 
	 * @param context
	 * @param net
	 * @param classType
	 * @return
	 */
	public static boolean createMarking(UIPluginContext context, PetrinetGraph net, Class<? extends Connection> classType) {
		boolean result = false;
		Collection<Pair<Integer, PluginParameterBinding>> plugins = context.getPluginManager().find(
				ConnectionObjectFactory.class, classType, context.getClass(), true, false, false, net.getClass());
		PluginContext c2 = context.createChildContext("Creating connection of Type " + classType);
		Pair<Integer, PluginParameterBinding> pair = plugins.iterator().next();
		PluginParameterBinding binding = pair.getSecond();
		try {
			PluginExecutionResult pluginResult = binding.invoke(c2, net);
			pluginResult.synchronize();
			context.getProvidedObjectManager().createProvidedObjects(c2); // push the objects to main context
			result = true;
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			c2.getParentContext().deleteChild(c2);
		}
		return result;
	}
	
	/**
	 * TODO: Write Javadoc
	 * 
	 * @param context
	 * @param net
	 * @return
	 */
	public static Marking getInitialMarking(PluginContext context, PetrinetGraph net) {
		// check connection between petri net and marking
		Marking initMarking = null;
		try {
			initMarking = context.getConnectionManager()
					.getFirstConnection(InitialMarkingConnection.class, context, net)
					.getObjectWithRole(InitialMarkingConnection.MARKING);
		} catch (ConnectionCannotBeObtained exc) {
			initMarking = new Marking();
		}
		return initMarking;
	}
	
	/**
	 * TODO: Write Javadoc
	 * 
	 * @param context
	 * @param net
	 * @return
	 */
	public static Marking[] getFinalMarkings(PluginContext context, PetrinetGraph net) {
		// check if final marking exists
		Marking[] finalMarkings = null;
		try {
			Collection<FinalMarkingConnection> finalMarkingConnections = context.getConnectionManager()
					.getConnections(FinalMarkingConnection.class, context, net);
			if (finalMarkingConnections.size() != 0) {
				Set<Marking> setFinalMarkings = new HashSet<Marking>();
				for (FinalMarkingConnection conn : finalMarkingConnections) {
					setFinalMarkings.add((Marking) conn.getObjectWithRole(FinalMarkingConnection.MARKING));
				}
				finalMarkings = setFinalMarkings.toArray(new Marking[setFinalMarkings.size()]);
			} else {
				finalMarkings = new Marking[0];
			}
		} catch (ConnectionCannotBeObtained exc) {
			// no final marking provided, give an empty marking
			finalMarkings = new Marking[0];
		}
		return finalMarkings;
	}
	
	/**
	 * TODO: Write Javadoc
	 * 
	 * @param net
	 * @param markingA
	 * @param markingB
	 * @return
	 */
	public static Marking mergeMarkings(PetriNetWithData net, Marking markingA, Marking markingB) {
		Marking mergedMarking = new Marking();
		Map<LocalNodeID,Place> netPlaces = getPlaceMap(net);
		
		for (Place place : markingA.baseSet()) {
			mergedMarking.add(netPlaces.get(place.getLocalID()), markingA.occurrences(place));
		}
		
		for (Place place : markingB.baseSet()) {
			mergedMarking.add(netPlaces.get(place.getLocalID()), markingB.occurrences(place));
		}
		
		return mergedMarking;
	}
	
	/**
	 * TODO: Write Javadoc
	 * 
	 * @param net
	 * @return
	 */
	public static Map<LocalNodeID,Place> getPlaceMap(PetriNetWithData net) {
		Map<LocalNodeID,Place> placeMap = new HashMap<LocalNodeID,Place>();
		
		for (Place place : net.getPlaces()) {
			placeMap.put(place.getLocalID(), place);
		}
		
		return placeMap;
	}
	
	/**
	 * TODO: Write Javadoc
	 * 
	 * @param net
	 * @return
	 */
	public static Map<LocalNodeID,Transition> getTransitionMap(PetriNetWithData net) {
		Map<LocalNodeID,Transition> transitionMap = new HashMap<LocalNodeID,Transition>();
		
		for (Transition transition : net.getTransitions()) {
			transitionMap.put(transition.getLocalID(), transition);
		}
		
		return transitionMap;
	}
}

