package org.processmining.datapetrinets.ui.create;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithData;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithDataFactory;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.widgets.wizard.Wizard;

public class ToPetriNetWithDataPlugin {
	@Plugin(name = "Create/Edit PetriNet With Data", level = PluginLevel.PeerReviewed, 
			parameterLabels = { "Petri Net with Reset and Inhibitor Arcs" }, 
			returnLabels = { "Petri Net with Data", "Initial Marking", "Final Marking" }, 
			returnTypes = { DataPetriNet.class,	Marking.class, Marking.class }, 
			userAccessible = true,
			keywords = {"DPN", "data", "editor" }, 
			help = "This plug-in can be used to modify Data Petri Nets, such as "+
			" adding/removing variables as well as modifying guards, write and read operations associated"
					+ " with transitions")
	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "M. de Leoni", email = "m.d.leoni@tue.nl")
	public static Object[] createPetriNet(UIPluginContext context, PetrinetGraph inputNet) {
		AddVariablesPanel varPanel;
		PetriNetWithDataFactory factory = new PetriNetWithDataFactory(inputNet, "New version of " + inputNet.getLabel());
		PetriNetWithData retValue = factory.getRetValue();
		Marking initialMarking=null,finalMarking=null;
		try
		{
			Marking markings[] = factory.cloneInitialAndFinalConnection(context);
			initialMarking = markings[0];
			finalMarking = markings[1];
			retValue.setInitialMarking(initialMarking);
			retValue.setFinalMarkings(new Marking[] { finalMarking });		
		}
		catch(Exception err)
		{
			context.log("Impossible to clone the initial and final markings: please redefine them!");
		}		
		if (inputNet instanceof DataPetriNet)
			varPanel = new AddVariablesPanel((DataPetriNet) inputNet,retValue,context,factory);
		else
			varPanel = new AddVariablesPanel(retValue,context,factory);
		
		if(Wizard.show(context, varPanel).getInteractionResult()==InteractionResult.CANCEL)
		{
			context.getFutureResult(0).cancel(true);
			return null;
		}
		else
			return new Object[] { retValue, initialMarking, finalMarking };

	}
}
