package org.processmining.datapetrinets.validation;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import javax.swing.event.TableModelListener;
import javax.swing.table.TableModel;

import org.cpntools.accesscpn.engine.highlevel.instance.Binding;
import org.cpntools.accesscpn.engine.highlevel.instance.Marking;
import org.cpntools.accesscpn.engine.highlevel.instance.State;
import org.processmining.framework.util.Pair;

public class DeadLockTableModel implements TableModel  {
	private ArrayList<Binding[]> executions=new ArrayList<Binding[]>();
	private String[][] tableValues; 
	private ArrayList<String> columnNames=new ArrayList<String>();
	private HashMap<String,Integer> nameToColumn=new HashMap<String,Integer>();
	
	public Binding[] getBinding(int index)
	{
		return executions.get(index);
	}

	public DeadLockTableModel(List<Pair<State, Binding[]>> deadlocks) 
	{
		if (deadlocks.size()>0)
		{
			int index=0;
			for(Marking m : deadlocks.iterator().next().getFirst().getAllMarkings())
			{
				String placeName=m.getPlaceInstance().getNode().getName().asString();
				if (!placeName.startsWith("Variable") && !placeName.startsWith("Potential"))
				{
					if (placeName.lastIndexOf('_')>=0)
						columnNames.add(placeName.substring(0, placeName.lastIndexOf('_')));
					else
						columnNames.add(placeName);
					nameToColumn.put(placeName, index++);
				}
			}
		}

		tableValues=new String[deadlocks.size()][columnNames.size()];

		int rowIndex=-1;
		for(Pair<State, Binding[]> pair : deadlocks)
		{
			rowIndex++;
			executions.add(pair.getSecond());
			for(Marking mark : pair.getFirst().getAllMarkings())
			{
				String placeName=mark.getPlaceInstance().getNode().getName().asString();
				Integer colIndex=nameToColumn.get(placeName);
				if (colIndex!=null)
				{
					if (mark.getMarking().trim().equals("empty"))
						tableValues[rowIndex][colIndex]="0";
					else
						tableValues[rowIndex][colIndex]=mark.getMarking().replace("`()", "");
				}
			}
		}
	}


	public Class<?> getColumnClass(int arg0) {
		return String.class;
	}

	public int getColumnCount() {
		return columnNames.size();
	}

	public String getColumnName(int arg0) {
		return columnNames.get(arg0);
	}

	public int getRowCount() {
		if (executions!=null)
			return executions.size();
		else
			return 0;
	}

	public Object getValueAt(int row, int column) {
		return tableValues[row][column];
	}

	public boolean isCellEditable(int arg0, int arg1) {
		return false;
	}


	public void addTableModelListener(TableModelListener l) {
		
	}


	public void removeTableModelListener(TableModelListener l) {
		
	}


	public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
		
	}
}
