package org.processmining.datapetrinets.visualization.graphviz;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dialog.ModalityType;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutionException;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextArea;
import javax.swing.KeyStroke;
import javax.swing.ListSelectionModel;
import javax.swing.SwingWorker;
import javax.swing.table.DefaultTableModel;

import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.utils.TikZUtil;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverter.DPNAsDot;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverter.DecorationKey;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverterPlugin.GuardDisplayMode;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverterPlugin.PlaceDisplayMode;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverterPlugin.VariableDisplayMode;
import org.processmining.framework.util.ui.widgets.ProMScrollPane;
import org.processmining.framework.util.ui.widgets.ProMSplitPane;
import org.processmining.framework.util.ui.widgets.ProMTable;
import org.processmining.framework.util.ui.widgets.ProMTextArea;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.models.graphbased.AbstractGraphElement;
import org.processmining.models.graphbased.directed.DirectedGraphNode;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.graphviz.dot.Dot;
import org.processmining.plugins.graphviz.dot.Dot.GraphDirection;
import org.processmining.plugins.graphviz.dot.DotElement;
import org.processmining.plugins.graphviz.dot.DotNode;
import org.processmining.plugins.graphviz.visualisation.DotPanel;
import org.processmining.plugins.graphviz.visualisation.listeners.DotElementSelectionListener;
import org.processmining.plugins.graphviz.visualisation.listeners.SelectionChangedListener;

import com.fluxicon.slickerbox.factory.SlickerFactory;
import com.google.common.base.Predicate;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import com.kitfox.svg.SVGDiagram;

public class DPNGraphvizVisualizer extends JPanel {

	private final class ShowTikzAction extends AbstractAction {

		private static final int DEFAULT_HEIGHT = 800;
		private static final int DEFAULT_WIDTH = 1000;

		private static final long serialVersionUID = 1L;

		private ShowTikzAction() {
			super("Show TikZ code");
		}

		public void actionPerformed(ActionEvent e) {
			TikZUtil tikZUtil = new TikZUtil();
			String tikzCode = tikZUtil.transformToTikZ(dpn, initialMarking != null ? initialMarking : new Marking(),
					finalMarkings != null ? finalMarkings[0] : new Marking());
			showCustom(new JTextArea(tikzCode), "TikZ code", ModalityType.MODELESS);
		}

		private void showCustom(JComponent component, String dialogTitle, ModalityType modalityType) {
			ProMScrollPane scrollPane = new ProMScrollPane(component);
			scrollPane.setBorder(BorderFactory.createLineBorder(Color.BLACK, 2));
			if (component.getPreferredSize().getHeight() > DEFAULT_HEIGHT
					|| component.getPreferredSize().getWidth() > DEFAULT_WIDTH) {
				scrollPane.setPreferredSize(new Dimension(DEFAULT_WIDTH, DEFAULT_HEIGHT));
			}
			scrollPane.setMaximumSize(new Dimension(DEFAULT_WIDTH, DEFAULT_HEIGHT));
			scrollPane.setAlignmentX(JComponent.TOP_ALIGNMENT);
			Object[] options = { "OK" };
			JOptionPane optionPane = new JOptionPane(scrollPane, JOptionPane.PLAIN_MESSAGE, JOptionPane.DEFAULT_OPTION,
					null, options, options[0]);
			JDialog dialog = optionPane.createDialog(dialogTitle);
			dialog.setResizable(true);
			dialog.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
			dialog.setModalityType(modalityType);
			dialog.pack();
			dialog.setVisible(true);
		}
	}

	//TODO change to a more structured way 

	private final class SelectionChangedListenerImplementation implements SelectionChangedListener<DotElement> {

		private final DPNAsDot dotConversionResult;

		private SelectionChangedListenerImplementation(DPNAsDot dotConversionResult) {
			this.dotConversionResult = dotConversionResult;
		}

		public void selectionChanged(Set<DotElement> selectedElements) {
			if (selectedElements.size() == 1) {
				DotElement element = selectedElements.iterator().next();
				PetrinetNode node = dotConversionResult.getMapping().inverse().get(element);
				if (node instanceof PNWDTransition) {
					showTransitionDetails(node, textArea);
				} else if (node instanceof Place) {
					showPlaceDetails(node, textArea);
				} else if (node instanceof DataElement) {
					showVariableDetails(node, textArea);
				} else {
					textArea.setText("No information available for this node.");
				}
			} else {
				textArea.setText("Click on a transition/place to get more information.");
			}
		}

	}

	private final class ElementSelectionListenerColorImpl implements DotElementSelectionListener {

		private final DotPanel dotPanel;

		private String defaultStrokeWidth;
		private String defaultStrokeDashArray;

		private ElementSelectionListenerColorImpl(DotPanel dotPanel, DotElement node) {
			this.dotPanel = dotPanel;
			defaultStrokeWidth = DotPanel.getAttributeOf(DotPanel.getSVGElementOf(dotPanel.getSVG(), node),
					"stroke-width");
			defaultStrokeDashArray = DotPanel.getAttributeOf(DotPanel.getSVGElementOf(dotPanel.getSVG(), node),
					"stroke-dasharray");
		}

		public void colorSelectedElement(final DotPanel dotPanel, DotElement element) {
			defaultStrokeWidth = DotPanel.setCSSAttributeOf(dotPanel.getSVG(), element, "stroke-width", "3");
			defaultStrokeDashArray = DotPanel.setCSSAttributeOf(dotPanel.getSVG(), element, "stroke-dasharray", "5,5");
		}

		public void selected(DotElement element, SVGDiagram image) {
			colorSelectedElement(dotPanel, element);
			dotPanel.repaint();
		}

		public void deselected(DotElement element, SVGDiagram image) {
			DotPanel.setCSSAttributeOf(dotPanel.getSVG(), element, "stroke-width", defaultStrokeWidth);
			DotPanel.setCSSAttributeOf(dotPanel.getSVG(), element, "stroke-dasharray", defaultStrokeDashArray);
			dotPanel.repaint();
		}
	}

	private static final long serialVersionUID = 1L;

	private final DataPetriNet dpn;
	private final Marking initialMarking;
	private final Marking[] finalMarkings;

	private final RatioAwareDotPanel dotPanel;
	private final DefaultTableModel tableModel;
	private final ProMTable table;
	private final ProMTextArea textArea;

	private VariableDisplayMode variableMode = VariableDisplayMode.AUTO_LAYOUT;
	private GuardDisplayMode guardMode = GuardDisplayMode.EDGES;
	private PlaceDisplayMode placeMode = PlaceDisplayMode.BASIC;
	private GraphDirection direction = GraphDirection.leftRight;

	private DPNGraphvizConverter converter = new DPNGraphvizConverter();

	public DPNGraphvizVisualizer(final DataPetriNet dpn, final Marking initialMarking, final Marking[] finalMarkings) {
		this.dpn = dpn;
		this.initialMarking = initialMarking;
		this.finalMarkings = finalMarkings;
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));

		ProMSplitPane splitPane = new ProMSplitPane(ProMSplitPane.VERTICAL_SPLIT);
		splitPane.setResizeWeight(1.0);
		splitPane.setOneTouchExpandable(true);

		textArea = new ProMTextArea(false);
		textArea.setText("Click on a transition/place to get more information.");
		textArea.setLineWrap(true);
		textArea.setMaximumSize(null);
		textArea.setMinimumSize(null);
		textArea.setPreferredSize(null);

		Dot dot = new Dot();
		dot.addNode("loading ...");
		dot.setDirection(direction);
		dotPanel = new RatioAwareDotPanel(dot);
		dotPanel.requestFocusInWindow();

		dotPanel.getHelperControlsExplanations().add("show TikZ code");
		dotPanel.getHelperControlsShortcuts().add("ctrl t");
		dotPanel.getInputMap(WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke(KeyEvent.VK_T, InputEvent.CTRL_MASK),
				"showTikz");
		dotPanel.getActionMap().put("showTikz", new ShowTikzAction());

		dotPanel.getHelperControlsExplanations().add("increase word-wrap limit");
		dotPanel.getHelperControlsShortcuts().add("ctrl o");
		dotPanel.getInputMap(WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke(KeyEvent.VK_O, InputEvent.CTRL_MASK),
				"increaseWordWrap");
		dotPanel.getActionMap().put("increaseWordWrap", new AbstractAction("Increase word-wrap limit") {

			private static final long serialVersionUID = -2268751444434867910L;

			public void actionPerformed(ActionEvent e) {
				converter.setMaxLabelLineLength(converter.getMaxLabelLineLength() + 1);
				updateDotModel();
			}
		});

		dotPanel.getHelperControlsExplanations().add("decrease word-wrap limit");
		dotPanel.getHelperControlsShortcuts().add("ctrl l");
		dotPanel.getInputMap(WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke(KeyEvent.VK_L, InputEvent.CTRL_MASK),
				"decreaseWordWrap");
		dotPanel.getActionMap().put("decreaseWordWrap", new AbstractAction("decrease word-wrap limit") {

			private static final long serialVersionUID = -2268751444434867910L;

			public void actionPerformed(ActionEvent e) {
				converter.setMaxLabelLineLength(Math.max(1, converter.getMaxLabelLineLength() - 1));
				updateDotModel();
			}
		});

		tableModel = createTableModel(dpn);
		table = new ProMTable(tableModel);
		table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		table.setPreferredSize(new Dimension(600, 200));

		splitPane.setTopComponent(dotPanel);

		ProMSplitPane bottomSplitpane = new ProMSplitPane(ProMSplitPane.HORIZONTAL_SPLIT);
		bottomSplitpane.setResizeWeight(1.0);
		bottomSplitpane.setOneTouchExpandable(true);
		bottomSplitpane.setLeftComponent(textArea);
		bottomSplitpane.setRightComponent(table);
		splitPane.setBottomComponent(bottomSplitpane);

		add(splitPane);

		JPanel bottomPanel = new JPanel();
		bottomPanel.setLayout(new BoxLayout(bottomPanel, BoxLayout.X_AXIS));

		@SuppressWarnings("rawtypes") // cannot change SlickerFactory
		final JComboBox variableCbx = SlickerFactory.instance().createComboBox(VariableDisplayMode.values());
		variableCbx.setMaximumSize(new Dimension(200, 30));
		variableCbx.setSelectedItem(VariableDisplayMode.AUTO_LAYOUT);
		variableCbx.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				variableMode = (VariableDisplayMode) variableCbx.getSelectedItem();
				updateDotModel();

			}
		});
		bottomPanel.add(createHeading("VARIABLE MODE"));
		bottomPanel.add(Box.createHorizontalStrut(5));
		bottomPanel.add(variableCbx);

		@SuppressWarnings("rawtypes") // cannot change SlickerFactory
		final JComboBox placeCbx = SlickerFactory.instance().createComboBox(PlaceDisplayMode.values());
		placeCbx.setMaximumSize(new Dimension(200, 30));
		placeCbx.setSelectedItem(PlaceDisplayMode.BASIC);
		placeCbx.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				placeMode = (PlaceDisplayMode) placeCbx.getSelectedItem();
				updateDotModel();

			}
		});
		bottomPanel.add(createHeading("PLACE MODE"));
		bottomPanel.add(Box.createHorizontalStrut(5));
		bottomPanel.add(placeCbx);

		@SuppressWarnings("rawtypes") // cannot change SlickerFactory
		final JComboBox guardCbx = SlickerFactory.instance().createComboBox(GuardDisplayMode.values());
		guardCbx.setMaximumSize(new Dimension(200, 30));
		guardCbx.setSelectedItem(GuardDisplayMode.EDGES);
		guardCbx.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				guardMode = (GuardDisplayMode) guardCbx.getSelectedItem();
				updateDotModel();

			}
		});
		bottomPanel.add(createHeading("GUARD MODE"));
		bottomPanel.add(Box.createHorizontalStrut(5));
		bottomPanel.add(guardCbx);

		add(bottomPanel);

		updateDotModel();
	}

	public void updateDotModel() {
		new SwingWorker<DPNAsDot, Void>() {

			protected DPNAsDot doInBackground() throws Exception {
				return converter.doConvertDPN(dpn, initialMarking, finalMarkings, guardMode, variableMode, placeMode,
						direction, Collections.<AbstractGraphElement, Map<DecorationKey, Object>>emptyMap());
			}

			protected void done() {
				try {
					final DPNAsDot dotConversionResult = get();
					dotPanel.changeDot(dotConversionResult.getDot(), true);
					for (final DotNode node : dotPanel.getDotPanel().getNodes()) {
						node.addSelectionListener(new ElementSelectionListenerColorImpl(dotPanel.getDotPanel(), node));
					}
					dotPanel.addSelectionChangedListener(
							new SelectionChangedListenerImplementation(dotConversionResult));
				} catch (InterruptedException | ExecutionException e) {
					ProMUIHelper.showErrorMessage(e.getMessage(), "Failed to compute DOT");
				}
			}

		}.execute();

	}

	private JLabel createHeading(String label) {
		JLabel title = new JLabel(label);
		title.setAlignmentX(Component.CENTER_ALIGNMENT);
		title.setFont(title.getFont().deriveFont(Font.BOLD));
		title.setForeground(null);
		return title;
	}

	protected void showPlaceDetails(DirectedGraphNode node, ProMTextArea textArea) {
		textArea.setText("Place: " + node.getLabel());
	}

	protected void showTransitionDetails(DirectedGraphNode node, ProMTextArea textArea) {
		PNWDTransition transition = (PNWDTransition) node;
		StringBuilder sb = new StringBuilder();
		sb.append("Transition: " + node.getLabel());
		if (transition.getGuardExpression() != null) {
			sb.append("\nGuard: " + transition.getGuardExpression().toPrettyString(1));
		}
		if (!transition.getWriteOperations().isEmpty()) {
			sb.append("\nWrites:");
			for (DataElement var : transition.getWriteOperations()) {
				sb.append("\n-" + var.getVarName());
			}
		}
		textArea.setText(sb.toString());
		textArea.setCaretPosition(0);
	}

	protected void showVariableDetails(PetrinetNode node, ProMTextArea textArea) {
		DataElement data = (DataElement) node;
		textArea.setText("Variable: " + data.getVarName() + " [" + data.getType().getSimpleName() + "]");
	}

	protected DefaultTableModel createTableModel(DataPetriNet dpn) {
		final List<Transition> transitions = getSortedTransitions(dpn);
		DefaultTableModel tableModel = new DefaultTableModel(new String[] { "Transition", "Guard" }, 0);
		for (Transition t : transitions) {
			String guard = ((PNWDTransition) t).getGuardAsString();
			tableModel.addRow(new String[] { t.getLabel(), guard });
		}
		return tableModel;
	}

	protected static List<Transition> getSortedTransitions(DataPetriNet dpn) {
		final List<Transition> transitions = Lists
				.newArrayList(Iterables.filter(dpn.getTransitions(), new Predicate<Transition>() {

					public boolean apply(Transition t) {
						return ((PNWDTransition) t).getGuardExpression() != null;
					}
				}));
		Collections.sort(transitions, new Comparator<Transition>() {

			public int compare(Transition o1, Transition o2) {
				return o1.getLabel() != null && o2.getLabel() != null ? o1.getLabel().compareTo(o2.getLabel())
						: o1.compareTo(o2);
			}
		});
		return transitions;
	}

	protected double caluclateAverageQualityMeasure(Collection<Transition> transitions) {
		double sum = 0;
		int count = 0;
		for (Transition t : transitions) {
			double quality = ((PNWDTransition) t).getQuality();
			if (quality > 0) {
				sum += quality;
				count++;
			}
		}
		if (count > 0) {
			return sum / count;
		} else {
			return 0;
		}
	}

	protected DataPetriNet getDpn() {
		return dpn;
	}

}
