package org.processmining.models.graphbased.directed.petrinetwithdata.newImpl;

import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.ListSelectionModel;
import javax.swing.table.DefaultTableModel;

import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.contexts.uitopia.annotations.Visualizer;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.util.ui.widgets.ProMSplitPane;
import org.processmining.framework.util.ui.widgets.ProMTable;
import org.processmining.framework.util.ui.widgets.ProMTextArea;
import org.processmining.models.graphbased.ViewSpecificAttributeMap;
import org.processmining.models.graphbased.directed.DirectedGraphNode;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.jgraph.ProMJGraphVisualizer;
import org.processmining.models.jgraph.visualization.ProMJGraphPanel;

import com.google.common.base.Predicate;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;

public class DPNVisualizer {

	private static class DPNVisualizerListener extends MouseAdapter {

		private final ProMJGraphPanel panel;
		private final PetriNetWithData net;
		private final ProMTextArea area;

		public DPNVisualizerListener(PluginContext context, ProMJGraphPanel panel, PetriNetWithData net,
				ProMTextArea area) {
			this.panel = panel;
			this.net = net;
			this.area = area;
		}

		@Override
		public void mouseClicked(MouseEvent arg0) {

			Collection<DirectedGraphNode> selectedNodes = panel.getSelectedNodes();
			if (!selectedNodes.isEmpty()) {
				DirectedGraphNode node = selectedNodes.iterator().next();
				if (node instanceof PNWDTransition) {
					showTransitionDetails(node);
				} else if (node instanceof Place && net instanceof DiscoveredPetriNetWithData) {
					showPlaceDetails(node);
				} else
					area.setText("No information available for this node.");
			} else {
				area.setText("Click on a transition/place to get more information.");
			}
		}

		@Override
		public void mouseReleased(MouseEvent e) {
			if (e.isPopupTrigger())
				showMenu(e);
		}

		@Override
		public void mousePressed(MouseEvent e) {
			if (e.isPopupTrigger())
				showMenu(e);
		}

		private void showMenu(MouseEvent e) {
			JPopupMenu menu = new JPopupMenu();

			Place selectedPlace = getSelectedPlace(panel.getSelectedNodes());
			if (selectedPlace != null) {
				JMenuItem menuDecisionTree = new JMenuItem("Discover Decision Rules for" + selectedPlace.getLabel());
				menuDecisionTree.addActionListener(new ActionListener() {

					public void actionPerformed(ActionEvent e) {

					}
				});
				menu.add(menuDecisionTree);
			}

			menu.show(e.getComponent(), e.getX(), e.getY());
		}

		private static Place getSelectedPlace(Collection<DirectedGraphNode> selectNodes) {
			Iterator<DirectedGraphNode> iterator = selectNodes.iterator();
			Place node = null;
			while (iterator.hasNext() && node == null) {
				DirectedGraphNode nextNode = iterator.next();
				if (nextNode instanceof Place) {
					return (Place) nextNode;
				}
			}
			return null;
		}

		public void showPlaceDetails(DirectedGraphNode node) {
			DecisionPointGuardsResult result = ((DiscoveredPetriNetWithData) net).getDecisionPointResult((Place) node);
			if (result != null)
				area.setText(result.toString() + "\n " + result.getQualityMeasureName() + ": "
						+ NumberFormat.getNumberInstance().format(result.getQualityMeasure()));
			else
				area.setText("No information available for this node.");
		}

		public void showTransitionDetails(DirectedGraphNode node) {
			PNWDTransition transition = (PNWDTransition) node;
			if (transition.getGuardExpression() != null) {
				String guard = transition.getGuardExpression().toString();
				double confidence = transition.getQuality();
				String multiLineGuard = guard;
				if (Double.compare(confidence, -1.0d) == 0) {
					area.setText("Transition: " + node.getLabel() + "\nGuard: " + multiLineGuard);
				} else {
					String qualityMeasureName = tryFindQualityMeasureName(transition);					
					area.setText("Transition: " + node.getLabel() + "\nGuard: " + multiLineGuard + "\n"
							+ qualityMeasureName + ": " + confidence);
				}
			} else {
				area.setText("Transition: " + node.getLabel() + "\nGuard: NO Guard");
			}
			area.setCaretPosition(0);
		}

		private String tryFindQualityMeasureName(PNWDTransition transition) {
			Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inEdges = transition
					.getGraph().getInEdges(transition);					
			if (!inEdges.isEmpty()) {
				// Get first place 
				for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge: inEdges) {
					if (edge.getSource() instanceof Place) {
						Place place = (Place) edge.getSource();
						return ((DiscoveredPetriNetWithData) net).getDecisionPointResult(place).getQualityMeasureName();
					}	
				}
			}
			return "F-Score";
		}

	}

	@Plugin(name = "Data Petri Net (JGraph)", parameterLabels = { "Data Petri Net (JGraph)" }, returnLabels = { "JComponent" }, returnTypes = { JComponent.class }, 
			userAccessible = true)
	@Visualizer
	@UITopiaVariant(affiliation = "TU/e", author = "Massimiliano", email = "m.d.leoni@tue.nl")
	public JComponent visualizePrediction(PluginContext context, PetriNetWithData dataPN) throws Exception {
		ProMSplitPane retValue = new ProMSplitPane(ProMSplitPane.VERTICAL_SPLIT);
		retValue.setResizeWeight(1.0);
		retValue.setOneTouchExpandable(true);

		final ProMJGraphPanel centerPanel = ProMJGraphVisualizer.instance().visualizeGraph(context, dataPN,
				new ViewSpecificAttributeMap());
		retValue.setTopComponent(centerPanel);

		ProMSplitPane bottomSplitpane = new ProMSplitPane(ProMSplitPane.HORIZONTAL_SPLIT);
		bottomSplitpane.setResizeWeight(1.0);
		bottomSplitpane.setOneTouchExpandable(true);

		ProMTextArea text = new ProMTextArea(false);
		text.setText("Click on a transition/place to get more information.");
		text.setLineWrap(true);
		bottomSplitpane.setLeftComponent(text);

		final DPNVisualizerListener visualizerListener = new DPNVisualizerListener(context, centerPanel, dataPN, text);
		centerPanel.getGraph().addMouseListener(visualizerListener);

		final List<Transition> transitions = getSortedTransitions(dataPN);
		DefaultTableModel tableModel;

		if (dataPN instanceof DiscoveredPetriNetWithData) {
			String qualityMeasureName = "F1-Score";
			if (!dataPN.getPlaces().isEmpty()) {
				// Just take the first quality measure name, as we assume all guards have been discovered with the same method here
				for (Place place: dataPN.getPlaces()) {
					DecisionPointGuardsResult decisionPointResult = ((DiscoveredPetriNetWithData) dataPN)
							.getDecisionPointResult(place);
					if (decisionPointResult != null) {
						qualityMeasureName = decisionPointResult.getQualityMeasureName();
						break;
					}					
				}
			}

			tableModel = new DefaultTableModel(new String[] { "Transition", qualityMeasureName, "Guard" }, 0);
			for (Transition t : transitions) {
				String guard = ((PNWDTransition) t).getGuardAsString();
				double confidence = ((PNWDTransition) t).getQuality();
				tableModel.addRow(new String[] { t.getLabel(), NumberFormat.getNumberInstance().format(confidence),
						guard });
			}
			tableModel.addRow(new String[] { "Average",
					NumberFormat.getNumberInstance().format(caluclateAverageQualityMeasure(dataPN.getTransitions())),
					"" });
		} else {
			tableModel = new DefaultTableModel(new String[] { "Transition", "Guard" }, 0);
			for (Transition t : transitions) {
				String guard = ((PNWDTransition) t).getGuardAsString();
				tableModel.addRow(new String[] { t.getLabel(), guard });
			}
		}

		ProMTable table = new ProMTable(tableModel);
		table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		table.setPreferredSize(new Dimension(600, 200));
		bottomSplitpane.setRightComponent(table);
		retValue.setBottomComponent(bottomSplitpane);

		return retValue;

	}

	private static List<Transition> getSortedTransitions(PetriNetWithData dataPN) {
		final ArrayList<Transition> transitions = Lists.newArrayList(Iterables.filter(dataPN.getTransitions(),
				new Predicate<Transition>() {

					public boolean apply(Transition t) {
						return ((PNWDTransition) t).getGuardExpression() != null;
					}
				}));
		Collections.sort(transitions, new Comparator<Transition>() {

			public int compare(Transition o1, Transition o2) {
				return o1.getLabel() != null && o2.getLabel() != null ? o1.getLabel().compareTo(o2.getLabel()) : o1
						.compareTo(o2);
			}
		});
		return transitions;
	}

	private double caluclateAverageQualityMeasure(Collection<Transition> transitions) {
		double sum = 0;
		int count = 0;
		for (Transition t : transitions) {
			double quality = ((PNWDTransition) t).getQuality();
			if (quality > 0) {
				sum += quality;
				count++;
			}
		}
		if (count > 0) {
			return sum / count;
		} else {
			return 0;
		}
	}

}
