package org.processmining.models.graphbased.directed.petrinetwithdata.newImpl;

import java.text.ParseException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.swing.SwingConstants;

import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.DataPetriNetsWithMarkings;
import org.processmining.datapetrinets.exception.NonExistingVariableException;
import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.directed.DirectedGraphEdge;
import org.processmining.models.graphbased.directed.DirectedGraphNode;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.ExpandableSubNet;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinet.impl.InhibitorNetImpl;
import org.processmining.models.guards.Expression;
import org.processmining.models.semantics.petrinet.Marking;

public class PetriNetWithData extends InhibitorNetImpl implements DataPetriNetsWithMarkings {

	public static Set<String> getAllVariableNames(DataPetriNet net) {
		Set<String> variableNames = new HashSet<>();
		for (DataElement element : net.getVariables()) {
			variableNames.add(element.getVarName());
		}
		return variableNames;
	}

	@SuppressWarnings("rawtypes")
	public static Map<String, Class> getVariableTypes(DataPetriNet net) {
		Map<String, Class> varType = new HashMap<>();
		for (DataElement element : net.getVariables()) {
			varType.put(element.getVarName(), element.getType());
		}
		return varType;
	}

	/**
	 * Character that is appended to a variable
	 */
	public final static char WRITE_CHAR = '\'';

	private Map<String, DataElement> variableSet = new HashMap<String, DataElement>();
	private Set<VariableAccess> varAccessArc = new HashSet<VariableAccess>();

	private Marking initialMarking = null;
	private Marking[] finalMarkings = null;

	public PetriNetWithData(String netName) {
		super(netName);
		getAttributeMap().put(AttributeMap.PREF_ORIENTATION, SwingConstants.WEST);
	}

	@Override
	public Set<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> getEdges() {
		Set<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> retValue = super.getEdges();
		retValue.addAll(varAccessArc);
		return retValue;
	}

	@Override
	public Set<PetrinetNode> getNodes() {
		Set<PetrinetNode> retValue = super.getNodes();
		retValue.addAll(variableSet.values());
		return retValue;
	}

	@SuppressWarnings("rawtypes")
	public DataElement addVariable(String varName, Class type, Comparable minValue, Comparable maxValue) {
		DataElement v = new DataElement(varName, type, minValue, maxValue, this);
		variableSet.put(v.getVarName(), v);
		graphElementAdded(v);
		return v;
	}

	public void removeAllVariablesNotInGuard() {
		Set<String> normalVariablesInGuards = new HashSet<>();
		Set<String> primeVariablesInGuards = new HashSet<>();
		for (Transition t : super.getTransitions()) {
			GuardExpression guard = ((PNWDTransition) t).getGuardExpression();
			if (guard != null) {
				normalVariablesInGuards.addAll(guard.getNormalVariables());
				primeVariablesInGuards.addAll(guard.getPrimeVariables());
			}

		}
		Iterator<Entry<String, DataElement>> varIterator = variableSet.entrySet().iterator();
		HashSet<DataElement> dataElemsToRemove = new HashSet<>();
		while (varIterator.hasNext()) {
			Entry<String, DataElement> entry = varIterator.next();
			if (!normalVariablesInGuards.contains(entry.getKey()) && !primeVariablesInGuards.contains(entry.getKey())) {
				dataElemsToRemove.add(entry.getValue());
				varIterator.remove();
			}
		}

		Iterator<VariableAccess> iter = varAccessArc.iterator();
		while (iter.hasNext()) {
			VariableAccess next = iter.next();
			if (dataElemsToRemove.contains(next.getSource())) {
				((PNWDTransition) next.getTarget()).removeReadOperation(this, ((PNWDTransition) next.getTarget()),
						(DataElement) next.getSource());
				iter.remove();
			}

			else if (dataElemsToRemove.contains(next.getTarget())) {
				((PNWDTransition) next.getSource()).removeWriteOperation(this, ((PNWDTransition) next.getSource()),
						(DataElement) next.getTarget());

				iter.remove();
			}
		}
	}

	public void assignReadOperation(Transition t, DataElement v) {
		if (!variableSet.values().contains(v))
			throw (new IllegalArgumentException("Undefined variable"));
		if (!transitions.contains(t))
			throw (new IllegalArgumentException("The transition does not belong to the model"));
		((PNWDTransition) t).assignReadOperation(this, t, v);
		VariableAccess arc = new VariableAccess(v, t);
		addVarAccessArc(arc);
	}

	private void addVarAccessArc(VariableAccess arc) {
		VariableAccess existingArc = null;
		Iterator<VariableAccess> iterator = varAccessArc.iterator();
		while (iterator.hasNext()) {
			existingArc = iterator.next();
			if (arc.getSource().equals(existingArc.getTarget()) && arc.getTarget().equals(existingArc.getSource()))
				break;
			existingArc = null;
		}
		if (existingArc != null) {
			existingArc.setBidirectional(true);
			graphElementChanged(existingArc);
		} else {
			if (varAccessArc.add(arc))
				graphElementAdded(arc);
		}
	}

	public void assignWriteOperation(Transition t, DataElement v) {
		if (!variableSet.values().contains(v))
			throw (new IllegalArgumentException("Undefined variable"));
		if (!transitions.contains(t))
			throw (new IllegalArgumentException("The transition does not belong to the model"));
		((PNWDTransition) t).assignWriteOperation(this, t, v);
		VariableAccess arc = new VariableAccess(t, v);
		addVarAccessArc(arc);
	}

	public void removeWriteOperation(Transition t, DataElement v) {
		((PNWDTransition) t).removeWriteOperation(this, t, v);
		Iterator<VariableAccess> iter = varAccessArc.iterator();
		while (iter.hasNext()) {
			VariableAccess variableAccess = iter.next();
			if (variableAccess.getSource().equals(t) && variableAccess.getTarget().equals(v)) {
				iter.remove();
			}
		}
	}

	public void removeReadOperation(Transition t, DataElement v) {
		((PNWDTransition) t).removeReadOperation(this, t, v);
		Iterator<VariableAccess> iter = varAccessArc.iterator();
		while (iter.hasNext()) {
			VariableAccess variableAccess = iter.next();
			if (variableAccess.getTarget().equals(t) && variableAccess.getSource().equals(v)) {
				iter.remove();
			}
		}
	}

	public Collection<DataElement> getVariables() {
		return Collections.unmodifiableCollection(variableSet.values());
	}

	public DataElement getVariable(String name) {
		return variableSet.get(name);
	}

	public void setGuard(Transition t, String guard) throws ParseException {
		try {
			((PNWDTransition) t).setGuard(this, guard);
		} catch (NonExistingVariableException | org.processmining.datapetrinets.expression.syntax.ParseException e) {
			//TODO change method to throw both exceptions 
			throw new ParseException(e.getMessage(), -1);
		}
	}

	public void setGuardFromString(Transition transition, String guardAsString)
			throws org.processmining.datapetrinets.expression.syntax.ParseException, NonExistingVariableException {
		((PNWDTransition) transition).setGuard(this, guardAsString);
	}

	public void setGuard(Transition transition, GuardExpression guard) throws NonExistingVariableException {
		((PNWDTransition) transition).setGuard(this, guard);
	}

	/**
	 * Use {@link #setGuard(Transition, GuardExpression)}!
	 */
	@Deprecated
	public void setGuard(Transition t, Expression guard) {
		((PNWDTransition) t).setGuard(this, guard);
	}

	@Override
	public synchronized Transition addTransition(String label, ExpandableSubNet parent) {
		Transition t = new PNWDTransition(label, this, parent);
		transitions.add(t);
		graphElementAdded(t);
		return t;
	}

	public void setGuard(Transition t, GuardExpression guard, double quality) throws NonExistingVariableException {
		((PNWDTransition) t).setGuard(this, guard, quality);
	}

	/**
	 * Use {@link #setGuard(Transition, GuardExpression, double)} instead
	 */
	@Deprecated
	public void setGuard(Transition t, Expression guard, double quality) {
		((PNWDTransition) t).setGuard(this, guard, quality);
	}

	@Override
	public synchronized void removeEdge(@SuppressWarnings("rawtypes") DirectedGraphEdge edge) {
		if (edge instanceof VariableAccess) {
			varAccessArc.remove(edge);
			graphElementRemoved(edge);
		} else {
			super.removeEdge(edge);
		}
	}

	@Override
	public synchronized void removeNode(DirectedGraphNode node) {
		if (node instanceof DataElement) {
			removeSurroundingEdges((DataElement)node);
			variableSet.remove(((DataElement) node).getVarName());			
			graphElementRemoved(node);
		} else {
			super.removeNode(node);
		}
	}


	@Override
	public synchronized Transition removeTransition(Transition transition) {
		for (Iterator<VariableAccess> iterator = varAccessArc.iterator(); iterator.hasNext();) {
			VariableAccess access = iterator.next();
			if (access.getSource().equals(transition) || access.getTarget().equals(transition)) {
				iterator.remove();
			}
		}
		return super.removeTransition(transition);
	}

	public Marking getInitialMarking() {
		return initialMarking;
	}

	public Marking[] getFinalMarkings() {
		return finalMarkings;
	}

	public void setInitialMarking(Marking marking) {
		initialMarking = marking;
	}

	public void setFinalMarkings(Marking[] markings) {
		finalMarkings = markings;
	}

	public void removeAllVariables() {
		for(Transition t : getTransitions())
		{
			try {
				setGuard(t, (String)null);
			} catch (ParseException e) {
				e.printStackTrace();
			}
			for(DataElement elem : getVariables())
			{
				removeReadOperation(t, elem);
				removeWriteOperation(t, elem);
			}
		}
		variableSet.clear();
	}

}
